/*
 * siusbctl.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <getopt.h>

#include "si1300.h"

/*
 * Clock
 */

struct siusb_clock {
    unsigned clock;
    unsigned val;
};

#define SIUSB_NR_CLOCK_VALUES   9

static struct siusb_clock siusb_clock_values[SIUSB_NR_CLOCK_VALUES] = {
    {20, SIUSB_CLK_20MHZ},
    {25, SIUSB_CLK_25MHZ},
    {30, SIUSB_CLK_30MHZ},
    {35, SIUSB_CLK_35MHZ},
    {40, SIUSB_CLK_40MHZ},
    {45, SIUSB_CLK_45MHZ},
    {50, SIUSB_CLK_50MHZ},
    {55, SIUSB_CLK_55MHZ},
    {60, SIUSB_CLK_60MHZ}
};

static unsigned val_to_clock(unsigned val)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++) {
        if (val == siusb_clock_values[i].val)
            return siusb_clock_values[i].clock;
    }
    fprintf(stderr, "unknown clock value\n");
    exit(1);
}

static unsigned clock_to_val(unsigned clock)
{
    unsigned i;

    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++) {
        if (clock == siusb_clock_values[i].clock)
            return siusb_clock_values[i].val;
    }
    fprintf(stderr, "bad clock, allowed are:\n");
    for (i = 0; i < SIUSB_NR_CLOCK_VALUES; i++)
        fprintf(stderr, "%u%s", siusb_clock_values[i].clock,
                i == SIUSB_NR_CLOCK_VALUES-1 ? "\n" : ", ");
    exit(1);
}

/*
 * Gain
 */

struct siusb_gain {
    double gain;
    unsigned val;
};

#define CLOCK_OPT           0
#define EXPOSURE_OPT        1
#define GAIN_OPT            2
#define NUMOPTS             3

static int option_present[NUMOPTS];
static const char *option_str[NUMOPTS];

static const struct option long_options[NUMOPTS+1] = {
    {"clock",         optional_argument, &option_present[CLOCK_OPT],         1},
    {"exposure",      optional_argument, &option_present[EXPOSURE_OPT],      1},
    {"gain",          optional_argument, &option_present[GAIN_OPT],          1},
    {NULL,            0,                 NULL,                               0}
};

static void set_clock(int fd)
{
    unsigned val;

    val = clock_to_val(atoi(option_str[CLOCK_OPT]));
    if (ioctl(fd, SIUSB_IOC_SET_CLOCK, &val) < 0) {
        perror("ioctl(SIUSB_IOC_SET_CLOCK)");
        exit(1);
    }
}

static void get_clock(int fd)
{
    unsigned val;

    if (ioctl(fd, SIUSB_IOC_GET_CLOCK, &val) < 0) {
        perror("ioctl(SIUSB_IOC_GET_CLOCK)");
        exit(1);
    }
    printf("clock = %u\n", val_to_clock(val));
}

static void set_exposure(int fd)
{
    unsigned val;

    val = atoi(option_str[EXPOSURE_OPT]);
    if (ioctl(fd, SIUSB_IOC_SET_EXPOSURE, &val) < 0) {
        perror("ioctl(SIUSB_IOC_SET_EXPOSURE)");
        exit(1);
    }
}

static void get_exposure(int fd)
{
    unsigned val;

    if (ioctl(fd, SIUSB_IOC_GET_EXPOSURE, &val) < 0) {
        perror("ioctl(SIUSB_IOC_GET_EXPOSURE)");
        exit(1);
    }
    printf("exposure = %u\n", val);
}

static void set_gain(int fd)
{
    unsigned val;

    val = atoi(option_str[GAIN_OPT]);
    if (ioctl(fd, SIUSB_IOC_SET_GAIN, &val) < 0) {
        perror("ioctl(SIUSB_IOC_SET_GAIN)");
        exit(1);
    }
}

static void get_gain(int fd)
{
    unsigned val;

    if (ioctl(fd, SIUSB_IOC_GET_GAIN, &val) < 0) {
        perror("ioctl(SIUSB_IOC_GET_GAIN)");
        exit(1);
    }
    printf("gain = %u\n", val);
}

static void usage(const char *argv0)
{
    fprintf(stderr, "\n");
    fprintf(stderr, "Usage: %s [OPTIONS] VIDEO_DEVICE\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, " -c, --clock[=CLOCK]                   get/set clock\n");
    fprintf(stderr, " -e, --exposure[=EXPOSURE]             get/set exposure\n");
    fprintf(stderr, " -g, --gain[=GAIN]                     get/set gain\n");
}

int main(int argc, char *argv[])
{
    int c;
    int option_index;
    char *dev;
    int fd;

    while (1) {
        c = getopt_long(argc, argv, "c::s::e::g::l::b::t::", long_options, &option_index);

        if (c == -1)
            break;

        switch (c) {
            case 0:     option_str[option_index] = optarg;
                        break;

            case 'c':   option_present[CLOCK_OPT] = 1;
                        option_str[CLOCK_OPT] = optarg;
                        break;

            case 'e':   option_present[EXPOSURE_OPT] = 1;
                        option_str[EXPOSURE_OPT] = optarg;
                        break;

            case 'g':   option_present[GAIN_OPT] = 1;
                        option_str[GAIN_OPT] = optarg;
                        break;

            case '?':   exit(1);    /* getopt_long() already printed an error message */
            default:    abort();    /* Should never happen */
        }
    }

    if (optind < argc)
        dev = argv[optind];
    else {
        usage(argv[0]);
        exit(1);
    }

    fd = open(dev, O_RDWR);
    if (fd < 0) {
        perror("open");
        exit(1);
    }

    if (option_present[CLOCK_OPT]) {
        if (option_str[CLOCK_OPT])
            set_clock(fd);
        else
            get_clock(fd);
    }

    if (option_present[EXPOSURE_OPT]) {
        if (option_str[EXPOSURE_OPT])
            set_exposure(fd);
        else
            get_exposure(fd);
    }

    if (option_present[GAIN_OPT]) {
        if (option_str[GAIN_OPT])
            set_gain(fd);
        else
            get_gain(fd);
    }

    /* If no options specified, print all.
     */

    for (option_index = 0; option_index < NUMOPTS; option_index++) {
        if (option_present[option_index])
            break;
    }

    if (option_index >= NUMOPTS) {
        get_clock(fd);
        get_exposure(fd);
        get_gain(fd);
    }

    close(fd);
    return 0;
}
