/*
 * siusb.c
 *
 * SI-1280 USB Camera Driver
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <linux/usb.h>
#include <asm/uaccess.h>
#include <linux/videodev.h>

#include "siusb.h"


/* Kernel compatibility
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define LINUX_24
#endif

#ifdef LINUX_24
#ifndef URB_ASYNC_UNLINK
#define URB_ASYNC_UNLINK    USB_ASYNC_UNLINK
#endif
#endif

#ifdef LINUX_24
/* 2.4 does not have the mem_flags argument */
#define usb_submit_urb(urb, mem_flags)  usb_submit_urb(urb)
#define usb_alloc_urb(urb, mem_flags)   usb_alloc_urb(urb)
#endif /* LINUX_24 */


/* Debug macros
 */
#ifdef DEBUG
#define TRACE1(fmt, args...)    do { if (DEBUG >= 1) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE2(fmt, args...)    do { if (DEBUG >= 2) printk(KERN_DEBUG fmt, ## args); } while (0)
#define TRACE3(fmt, args...)    do { if (DEBUG >= 3) printk(KERN_DEBUG fmt, ## args); } while (0)
#else /* !DEBUG */
#define TRACE1(fmt, args...)    do { } while (0)
#define TRACE2(fmt, args...)    do { } while (0)
#define TRACE3(fmt, args...)    do { } while (0)
#endif /* !DEBUG */


#define DRIVER_AUTHOR   "Nikka Research Deutschland GmbH"
#define DRIVER_DESC     "SI-1280 USB Camera Driver"
#define DRIVER_VERSION  "v0.5"


/* Module parameters
 */

#ifndef SIUSB_NUM_BUFS
#define SIUSB_NUM_BUFS      8
#endif

static unsigned num_bufs = SIUSB_NUM_BUFS;

MODULE_PARM(num_bufs, "i");
MODULE_PARM_DESC(num_bufs,
    "Number of buffers (default is " __stringify(SIUSB_NUM_BUFS) ")");

#ifndef SIUSB_NUM_BULK_URBS
#define SIUSB_NUM_BULK_URBS 32
#endif

static unsigned num_bulk_urbs = SIUSB_NUM_BULK_URBS;

MODULE_PARM(num_bulk_urbs, "i");
MODULE_PARM_DESC(num_bulk_urbs,
    "Number of bulk URBs (default is " __stringify(SIUSB_NUM_BULK_URBS) ")");


#define SIUSB_PACKET_SIZE   512

#define SIUSB_VERSION_SIZE  16


#define SIUSB_MIN_WIDTH     256
#define SIUSB_MIN_HEIGHT    1

#define SIUSB_MAX_WIDTH     1280
#define SIUSB_MAX_HEIGHT    1024

#define SIUSB_MAX_SIZE      (SIUSB_MAX_WIDTH * SIUSB_MAX_HEIGHT)


#define SIUSB_INIT_URB_TIMEOUT  (HZ/2)
#define SIUSB_BULK_URB_TIMEOUT  (HZ/2)

#define SIUSB_CTRL_TIMEOUT      (5*HZ)  /* Control transfer timeout in jiffies */


#ifndef VID_HARDWARE_SI
#define VID_HARDWARE_SI     71232   /* Random */
#endif


#define SIUSB_BUF_UNUSED    0   /* Unused (no MCAPTURE) */
#define SIUSB_BUF_GRABBING  1   /* In the process of being grabbed into */
#define SIUSB_BUF_DONE      2   /* Finished grabbing, but not been synced yet */

struct siusb_buf {
    unsigned nr;                /* Index in buffers array */
    unsigned state;             /* State of grabbing */
    int result;                 /* Negative error code or 0 = no error */
    unsigned long *pages;       /* Pages array */
    wait_queue_head_t done;     /* Wait for readout finish */
    struct list_head captq;     /* Capture queue list */
};

struct siusb {
    struct video_device *vdev;  /* V4L device for that camera */
    struct usb_device *udev;    /* Save USB device pointer */
    int open_count;             /* Device open count */
    int present;                /* True if device is not disconnected */
    struct semaphore sem;       /* Locks this structure */
    int bulk_ep_addr;           /* Bulk IN endpoint address */

    char version[SIUSB_VERSION_SIZE];   /* Keep camera version string */

    struct siusb_buf *bufs;     /* Buffers array */

    struct list_head captq;     /* Capture queue */
    spinlock_t captq_lock;      /* Lock capture queue add/remove */

    struct urb *init_urb;       /* Used to initiate readout */
    struct urb **bulk_urbs;     /* Queue of bulk transfer requests */

    int nr_req;                 /* Number of queued bulk transfer requests */
    int rem;                    /* Remaining bytes */
    unsigned cpage;             /* Current page in pages array */

    unsigned x, y, w, h;        /* Current window origin and size */
    unsigned shutter_mode;      /* Current shutter mode */
    unsigned exposure;          /* Current exposure */
    unsigned gain;              /* Current gain */
    unsigned black_level;       /* Current black level */
    unsigned black_balance;     /* Current black balance */

    unsigned trig_mode;         /* Trigger mode */
};


static unsigned siusb_pages_per_buf;    /* Numer of pages in pages array of each buffer */
static unsigned siusb_read_buf_nr;      /* read() syscall uses this buf */

static DECLARE_MUTEX(disconnect_sem);   /* Prevent races between open() and disconnect() */


/*
 * Low-level
 */

/* Vendor request codes
 */
#define SIUSB_REQ_VERSION           0xE1    /* Read firmware version */
#define SIUSB_REQ_FLASH             0xE4    /* Read/Write flash memory */
#define SIUSB_REQ_CLOCK             0xE2    /* Get/Set clock */
#define SIUSB_REQ_LOAD_REG          0xEF    /* Load sensor registers */
#define SIUSB_REQ_SET_ORIGIN        0xE6    /* Set row/column origin (pan/tilt) */
#define SIUSB_REQ_SET_DEPTH         0xED    /* Set bit depth (8 or 16 bits) */
#define SIUSB_REQ_READOUT           0xEA    /* Snapshot expose & readout */
#define SIUSB_REQ_ARM_TRIG          0xEC    /* Arm triggered, snapshot exposure mode */
#define SIUSB_REQ_DISARM_TRIG       0xEB    /* Disarm triggered mode */

/* Registers
 */
#define SIUSB_REG_SHUTTER_MODE      0x00
#define SIUSB_REG_WIDTH             0x01    /* (Pixels in row) / 2 - 1 */
#define SIUSB_REG_HEIGHT            0x02    /* Number of rows - 1 */
#define SIUSB_REG_EXPOSURE_TIME     0x03
#define SIUSB_REG_BEGIN_COLUMN      0x04    /* X-origin / 2 */
#define SIUSB_REG_BEGIN_ROW         0x05    /* Y-origin */
#define SIUSB_REG_BEGIN_ROW_2       0x06    /* Set to same value as Reg5 */
#define SIUSB_REG_SUBSAMPLING       0x07
#define SIUSB_REG_GAIN              0x08
#define SIUSB_REG_BLACK_LEVEL       0x09
#define SIUSB_REG_BLACK_BALANCE     0x0A

/* Bit depth
 */
#define SIUSB_DEPTH_8BIT    0
#define SIUSB_DEPTH_16BIT   1

static int siusb_ctrl_msg(struct siusb *s, __u8 dir, __u8 request,
                          __u16 value, __u16 index, void *data, __u16 size)
{
    return usb_control_msg(s->udev,
                           dir ? usb_rcvctrlpipe(s->udev, 0) : usb_sndctrlpipe(s->udev, 0),
                           request, dir | USB_TYPE_VENDOR | USB_RECIP_ENDPOINT,
                           value, index, data, size, SIUSB_CTRL_TIMEOUT);
}

static int siusb_read_version(struct siusb *s, char *buf)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_VERSION, 0, 0, buf, SIUSB_VERSION_SIZE);
    if (result != SIUSB_VERSION_SIZE) {
        printk(KERN_ERR "siusb: %s()=%d: FAILED\n", __func__, result);
        if (result >= 0)
            result = -EIO;
    } else {
        /* Be sure we return with properly null-terminated buf.
         */
        buf[SIUSB_VERSION_SIZE-1] = '\0';
    }
    return result;
}

static int siusb_get_clock(struct siusb *s, unsigned *clk)
{
    int result;

    /* The following transfer sets only the first 3 bytes of the passed buffer,
     * set the fourth byte to 0.
     */
    *clk = 0;
    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_CLOCK, 0, 0, clk, 3);
    if (result != 3) {
        printk(KERN_ERR "siusb: %s()=%d: FAILED\n", __func__, result);
        if (result >= 0)
            result = -EIO;
    }
    return result;
}

static int siusb_set_clock(struct siusb *s, unsigned clk)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_OUT, SIUSB_REQ_CLOCK, 0, 0, &clk, 3);
    if (result != 3) {
        printk(KERN_ERR "siusb: %s(clk=0x%08X)=%d: FAILED\n", __func__, clk, result);
        if (result >= 0)
            result = -EIO;
    }
    return result;
}

static int siusb_load_reg(struct siusb *s, __u16 value, __u16 index)
{
    return siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_LOAD_REG, value, index, NULL, 0);
}

static int siusb_set_origin(struct siusb *s, unsigned x, unsigned y)
{
    int result;

    x = (x/2)*2;
    if (x + SIUSB_MIN_WIDTH > SIUSB_MAX_WIDTH || y + SIUSB_MIN_HEIGHT > SIUSB_MAX_HEIGHT)
        return -EINVAL;
    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_SET_ORIGIN, x, y, NULL, 0);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(x=%u,y=%u)=%d: FAILED\n", __func__, x, y, result);
    else {
        s->x = x;
        s->y = y;
    }
    return result;
}

static int siusb_set_depth(struct siusb *s, int depth)
{
    int result;

    result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_SET_DEPTH, depth, 0, NULL, 0);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(depth=%d)=%d: FAILED\n", __func__, depth, result);
    return result;
}

static int siusb_set_trig_mode(struct siusb *s, unsigned mode)
{
    int result;

    if (mode)
        result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_ARM_TRIG, 0, 0, NULL, 0);
    else
        result = siusb_ctrl_msg(s, USB_DIR_IN, SIUSB_REQ_DISARM_TRIG, 0, 0, NULL, 0);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(mode=%u)=%d: FAILED\n", __func__, mode, result);
    else
        s->trig_mode = mode;
    return result;
}

static int siusb_set_shutter_mode(struct siusb *s, unsigned mode)
{
    int result;

    result = siusb_load_reg(s, mode, SIUSB_REG_SHUTTER_MODE);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(mode=0x%04X)=%d: FAILED\n", __func__, mode, result);
    else
        s->shutter_mode = mode;
    return result;
}

static int siusb_set_width(struct siusb *s, unsigned w)
{
    int result;

    w = (w/2)*2;
    if (w < SIUSB_MIN_WIDTH)
        return -EINVAL;
    if (w > SIUSB_MAX_WIDTH)
        w = SIUSB_MAX_WIDTH;
    result = siusb_load_reg(s, w/2-1, SIUSB_REG_WIDTH);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(w=%u)=%d: FAILED\n", __func__, w, result);
    else
        s->w = w;
    return result;
}

static int siusb_set_height(struct siusb *s, unsigned h)
{
    int result;

    if (h < SIUSB_MIN_HEIGHT)
        return -EINVAL;
    if (h > SIUSB_MAX_HEIGHT)
        h = SIUSB_MAX_HEIGHT;
    result = siusb_load_reg(s, h-1, SIUSB_REG_HEIGHT);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(h=%u)=%d: FAILED\n", __func__, h, result);
    else
        s->h = h;
    return result;
}

/* TODO: understand how this works and what exp actually means for the
 * different TBs.
 */
/*
 * TODO:
 *
 *  * Shutter mode = TRIG_TB3 -- the first image line is bad. This seems not to
 *    happen on TB4.
 */
static int siusb_set_exposure(struct siusb *s, unsigned exp)
{
    int result;

    result = siusb_load_reg(s, exp, SIUSB_REG_EXPOSURE_TIME);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(exp=0x%04X)=%d: FAILED\n", __func__, exp, result);
    else
        s->exposure = exp;
    return result;
}

static int siusb_set_gain(struct siusb *s, unsigned gain)
{
    int result;

    if (gain < SIUSB_GAIN_MIN || gain > SIUSB_GAIN_UNITY)
        return -EINVAL;
    result = siusb_load_reg(s, gain, SIUSB_REG_GAIN);
    if (result < 0)
        printk(KERN_ERR "siusb: %s(gain=0x%02X)=%d: FAILED\n", __func__, gain, result);
    else
        s->gain = gain;
    return result;
}

static int siusb_set_black_level(struct siusb *s, unsigned val)
{
    int result;

    if (val > SIUSB_BLACK_LEVEL_MAX)
        return -EINVAL;
    result = siusb_load_reg(s, val, SIUSB_REG_BLACK_LEVEL);
    if (result < 0)
        printk("siusb: %s(val=%u)=%d: FAILED\n", __func__, val, result);
    else
        s->black_level = val;
    return result;
}

static int siusb_set_black_balance(struct siusb *s, unsigned val)
{
    int result;

    if (val > SIUSB_BLACK_BALANCE_MAX)
        return -EINVAL;
    result = siusb_load_reg(s, val, SIUSB_REG_BLACK_BALANCE);
    if (result < 0)
        printk("siusb: %s(val=%u)=%d: FAILED\n", __func__, val, result);
    else
        s->black_balance = val;
    return result;
}


/*
 * Capture queue
 */

static int siusb_submit_bulk_urb(struct siusb *s, struct urb *urb, struct siusb_buf *buf);
static void siusb_readout_start(struct siusb *s);

static void siusb_readout_complete(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    unsigned start_readout = 0;

    /* Frame transfer is complete, when all queued URBs are complete.
     */
    if (s->nr_req > 0)
        return;

    TRACE2("siusb: buf#%u: DONE, result=%d\n", buf->nr, buf->result);

    spin_lock_irqsave(&s->captq_lock, flags);

    list_del(&buf->captq);

    if (!list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start readout\n", __func__, buf->nr);
        start_readout = 1;
    }

    spin_unlock_irqrestore(&s->captq_lock, flags);

    buf->state = SIUSB_BUF_DONE;
    TRACE2("siusb: buf#%u: wakeing up\n", buf->nr);
    wake_up(&buf->done);

    if (start_readout)
        siusb_readout_start(s);
}

static void siusb_report_urb_status(struct siusb *s, struct siusb_buf *buf, struct urb *urb)
{
    TRACE3("siusb: buf#%u: %s %p: status=%d, req %d, rcv %d\n", buf->nr,
           urb == s->init_urb ? "INIT" : "BULK", urb, urb->status,
           urb->transfer_buffer_length, urb->actual_length);
    if (urb->status < 0) {
        printk(KERN_ERR "siusb: buf#%u: %s %p: status=%d, req %d, rcv %d: FAILED\n", buf->nr,
               urb == s->init_urb ? "INIT" : "BULK", urb, urb->status,
               urb->transfer_buffer_length, urb->actual_length);
    }
}

#ifdef LINUX_24
static void siusb_init_urb_cb(struct urb *urb)
#else
static void siusb_init_urb_cb(struct urb *urb, struct pt_regs *regs)
#endif
{
    struct siusb *s = urb->context;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);
    unsigned i;

    siusb_report_urb_status(s, buf, urb);

    buf->result = urb->status;
    if (buf->result < 0)
        goto complete;

    for (i = 0; i < num_bulk_urbs; i++) {
        if (siusb_submit_bulk_urb(s, s->bulk_urbs[i], buf))
            goto complete;
    }

    return;

complete:
    siusb_readout_complete(s, buf);
}

#ifdef LINUX_24
static void siusb_bulk_urb_cb(struct urb *urb)
#else
static void siusb_bulk_urb_cb(struct urb *urb, struct pt_regs *regs)
#endif
{
    struct siusb *s = urb->context;
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    siusb_report_urb_status(s, buf, urb);

    s->nr_req--;

    if (buf->result < 0)
        goto complete;

    buf->result = urb->status;
    if (buf->result < 0)
        goto complete;

    if (siusb_submit_bulk_urb(s, urb, buf))
        goto complete;

    return;

complete:
    siusb_readout_complete(s, buf);
}

/* Returns 1 if transfer should be completed.
 */
static int siusb_submit_bulk_urb(struct siusb *s, struct urb *urb, struct siusb_buf *buf)
{
    if (s->rem <= 0)
        return 1;

    urb->transfer_buffer = (void *)buf->pages[s->cpage];
    urb->transfer_buffer_length = min(s->rem, (int)PAGE_SIZE);
    buf->result = usb_submit_urb(urb, GFP_ATOMIC);
    if (buf->result < 0) {
        printk(KERN_ERR "siusb: %s: buf#%u: usb_submit_urb()=%d: FAILED\n", __func__, buf->nr,
               buf->result);
        return 1;
    }
    TRACE2("siusb: buf#%u: BULK %p submitted, length=%d, rem=%d\n", buf->nr,
           urb, urb->transfer_buffer_length, s->rem);

    s->cpage++;
    s->rem -= PAGE_SIZE;

    s->nr_req++;
    return 0;
}

static void siusb_readout_start(struct siusb *s)
{
    struct siusb_buf *buf = list_entry(s->captq.next, struct siusb_buf, captq);

    s->nr_req = 0;

    /* Packet-size aligned remaining bytes.
     */
    s->rem = (s->w*s->h+SIUSB_PACKET_SIZE-1) & ~(SIUSB_PACKET_SIZE-1);

    s->cpage = 0;

    /* Submit init URB.
     */
    s->init_urb->transfer_buffer = (void *)buf->pages[0];
    buf->result = usb_submit_urb(s->init_urb, GFP_ATOMIC);
    if (buf->result < 0) {
        printk(KERN_ERR "siusb: %s: buf#%u: usb_submit_urb()=%d: FAILED\n", __func__, buf->nr,
               buf->result);
        siusb_readout_complete(s, buf);
        return;
    }
    TRACE2("siusb: buf#%u: INIT submitted\n", buf->nr);
}

static void __siusb_readout_wait(struct siusb *s, struct siusb_buf *buf)
{
    TRACE2("siusb: buf#%u: %d (%s): going to sleep\n", buf->nr, current->pid, current->comm);
    wait_event(buf->done, buf->state == SIUSB_BUF_DONE);
    TRACE2("siusb: buf#%u: %d (%s): awoken\n", buf->nr, current->pid, current->comm);
}

static int siusb_readout_wait(struct siusb *s, struct siusb_buf *buf)
{
    if (buf->state == SIUSB_BUF_UNUSED)
        return -EINVAL;
    __siusb_readout_wait(s, buf);
    buf->state = SIUSB_BUF_UNUSED;
    return 0;
}

static void siusb_flush_captq(struct siusb *s)
{
    struct siusb_buf *buf;

    /* Wait for the last (most recently) queued capture request to complete.
     */
    if (!list_empty(&s->captq)) {
        buf = list_entry(s->captq.prev, struct siusb_buf, captq);

        TRACE1("siusb: %s\n", __func__);
        __siusb_readout_wait(s, buf);
    }
}

static int siusb_capture_request(struct siusb *s, struct siusb_buf *buf)
{
    unsigned long flags;
    int start_readout = 0;

    if (buf->state != SIUSB_BUF_UNUSED)
        return -EBUSY;
    buf->state = SIUSB_BUF_GRABBING;

    spin_lock_irqsave(&s->captq_lock, flags);

    /* If there's no readout currently running (i.e. capture queue was empty
     * until now), then we must start one.
     */
    if (list_empty(&s->captq)) {
        TRACE2("siusb: %s: buf#%u: going to start readout\n", __func__, buf->nr);
        start_readout = 1;
    }

    list_add_tail(&buf->captq, &s->captq);

    spin_unlock_irqrestore(&s->captq_lock, flags);

    if (start_readout)
        siusb_readout_start(s);
    return 0;
}


/*
 * Resource allocation
 */

static void siusb_free_bufs(struct siusb *s)
{
    unsigned i, j;
    struct siusb_buf *buf;

    if (s->bufs) {
        for (i = 0; i < num_bufs; i++) {
            buf = &s->bufs[i];

            if (buf->pages) {
                for (j = 0; j < siusb_pages_per_buf; j++) {
                    if (buf->pages[j])
                        free_page(buf->pages[j]);
                }

                kfree(buf->pages);
            }
        }

        kfree(s->bufs);
        s->bufs = NULL;

        TRACE1("siusb: buffers released\n");
    }
}

static int siusb_alloc_bufs(struct siusb *s)
{
    unsigned i, j;
    struct siusb_buf *buf;

    s->bufs = kmalloc(num_bufs * sizeof(*s->bufs), GFP_KERNEL);
    if (!s->bufs) {
        printk(KERN_ERR "siusb: kmalloc(bufs): FAILED\n");
        goto err;
    }

    memset(s->bufs, 0, num_bufs * sizeof(*s->bufs));

    for (i = 0; i < num_bufs; i++) {
        buf = &s->bufs[i];

        buf->nr = i;
        buf->state = SIUSB_BUF_UNUSED;
        init_waitqueue_head(&buf->done);

        buf->pages = kmalloc(siusb_pages_per_buf * sizeof(*buf->pages), GFP_KERNEL);
        if (!buf->pages) {
            printk(KERN_ERR "siusb: buf#%u: kmalloc(pages): FAILED\n", buf->nr);
            goto err;
        }

        memset(buf->pages, 0, siusb_pages_per_buf * sizeof(*buf->pages));

        for (j = 0; j < siusb_pages_per_buf; j++) {
            buf->pages[j] = __get_free_page(GFP_KERNEL);
            if (!buf->pages[j]) {
                printk(KERN_ERR "siusb: buf#%u: __get_free_page(pages[%u]): FAILED\n", buf->nr, j);
                goto err;
            }
        }
    }

    TRACE1("siusb: buffers allocated\n");
    return 0;

err:
    siusb_free_bufs(s);
    return -ENOMEM;
}

static void siusb_free_urbs(struct siusb *s)
{
    unsigned i;

    if (s->bulk_urbs) {
        for (i = 0; i < num_bulk_urbs; i++) {
            if (s->bulk_urbs[i])
                usb_free_urb(s->bulk_urbs[i]);
        }

        kfree(s->bulk_urbs);
        s->bulk_urbs = NULL;
    }

    if (s->init_urb) {
        usb_free_urb(s->init_urb);
        s->init_urb = NULL;
    }
}

static int siusb_alloc_urbs(struct siusb *s)
{
    unsigned i;

    /* Allocate and initialize init URB.
     */

    s->init_urb = usb_alloc_urb(0, GFP_KERNEL);
    if (!s->init_urb) {
        printk(KERN_ERR "siusb: usb_alloc_urb(init_urb): FAILED\n");
        goto err;
    }

#ifdef LINUX_24
    s->init_urb->transfer_flags = URB_ASYNC_UNLINK;
    s->init_urb->timeout = SIUSB_INIT_URB_TIMEOUT;
#else
    usb_unlink_urb(s->init_urb);
#endif
    usb_fill_bulk_urb(s->init_urb, s->udev, usb_rcvbulkpipe(s->udev, s->bulk_ep_addr),
                      NULL, 2, siusb_init_urb_cb, s);

    /* Allocate and initialize bulk URBs.
     */

    s->bulk_urbs = kmalloc(num_bulk_urbs * sizeof(*s->bulk_urbs), GFP_KERNEL);
    if (!s->bulk_urbs) {
        printk(KERN_ERR "siusb: kmalloc(bulk_urbs): FAILED\n");
        goto err;
    }

    memset(s->bulk_urbs, 0, num_bulk_urbs * sizeof(*s->bulk_urbs));

    for (i = 0; i < num_bulk_urbs; i++) {
        s->bulk_urbs[i] = usb_alloc_urb(0, GFP_KERNEL);
        if (!s->bulk_urbs[i]) {
            printk(KERN_ERR "siusb: usb_alloc_urb(bulk_urbs[%u]): FAILED\n", i);
            goto err;
        }

#ifdef LINUX_24
        s->bulk_urbs[i]->transfer_flags = URB_ASYNC_UNLINK | URB_SHORT_NOT_OK;
        s->bulk_urbs[i]->timeout = SIUSB_BULK_URB_TIMEOUT;
#else
        usb_unlink_urb(s->bulk_urbs[i]);
#endif
        usb_fill_bulk_urb(s->bulk_urbs[i], s->udev, usb_rcvbulkpipe(s->udev, s->bulk_ep_addr),
                          NULL, 0, siusb_bulk_urb_cb, s);
    }

    return 0;

err:
    siusb_free_urbs(s);
    return -ENOMEM;
}


/*
 * V4L interface
 */

static int siusb_set_size(struct siusb *s, unsigned x, unsigned y, unsigned w, unsigned h)
{
    int result;

    TRACE1("siusb: %s: x=%u,y=%u,w=%u,h=%u\n", __func__, x, y, w, h);

    if (x + w > SIUSB_MAX_WIDTH || y + h > SIUSB_MAX_HEIGHT)
        return -EINVAL;

    /* Wait for all capture requests to complete before changing frame size.
     */
    siusb_flush_captq(s);

    result = siusb_set_origin(s, x, y);
    if (result < 0)
        return result;

    result = siusb_set_width(s, w);
    if (result < 0)
        return result;

    result = siusb_set_height(s, h);
    if (result < 0)
        return result;

    return 0;
}

static int siusb_open(struct inode *inode, struct file *file)
{
    struct video_device *vdev;
    struct siusb *s;
    int result;

    result = down_interruptible(&disconnect_sem);   /* Prevent disconnects */
    if (result < 0)
        return result;

    vdev = video_devdata(file);
    if (!vdev) {
        up(&disconnect_sem);
        return -ENODEV;
    }

    s = video_get_drvdata(vdev);

    result = down_interruptible(&s->sem);
    if (result < 0) {
        up(&disconnect_sem);
        return result;
    }

    if (s->open_count <= 0) {
        /* First open().
         */
        result = siusb_alloc_bufs(s);   /* Allocate image buffers */
        if (result < 0)
            goto out;
        result = siusb_alloc_urbs(s);   /* Allocate URBs */
        if (result < 0) {
            siusb_free_bufs(s);
            goto out;
        }
    } else {
        /* Allow single open.
         */
        result = -EBUSY;
        goto out;
    }

    s->open_count++;
    file->private_data = s;

out:
    up(&s->sem);
    up(&disconnect_sem);
    return result;
}

static int siusb_release(struct inode *inode, struct file *file)
{
    struct siusb *s = file->private_data;
    int result = 0;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    siusb_flush_captq(s);

    s->open_count--;

    if (s->open_count <= 0) {
        /* Last release().
         */
        siusb_free_urbs(s);
        siusb_free_bufs(s);
    }

    if (!s->present) {
        /* The camera was unplugged before the file was released.
         */
        up(&s->sem);
        kfree(s);
        return 0;
    }

    up(&s->sem);
    return result;
}

static ssize_t siusb_read(struct file *file, char *ptr, size_t count, loff_t *ppos)
{
    struct siusb *s = file->private_data;
    int rem;
    unsigned cpage;
    struct siusb_buf *buf = &s->bufs[siusb_read_buf_nr];
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    result = siusb_capture_request(s, buf);
    if (result < 0)
        goto out;

    result = siusb_readout_wait(s, buf);
    BUG_ON(result < 0);

    if (buf->result < 0) {
        result = buf->result;
        goto out;
    }

    if (count > s->w * s->h)
        count = s->w * s->h;

    for (rem = count, cpage = 0; rem > 0; rem -= PAGE_SIZE, cpage++) {
        if (copy_to_user(ptr, (void *)buf->pages[cpage], min(rem, (int)PAGE_SIZE))) {
            result = -EFAULT;
            goto out;
        }

        ptr += PAGE_SIZE;
    }

    result = count;

out:
    up(&s->sem);
    return result;
}

static unsigned siusb_poll(struct file *file, struct poll_table_struct *wait)
{
    struct siusb *s = file->private_data;
    struct siusb_buf *buf = &s->bufs[siusb_read_buf_nr];

    /* TODO: any way to use poll for ALL buffers, not only for read() call?
     */

    poll_wait(file, &buf->done, wait);
    if (buf->state == SIUSB_BUF_DONE) {
        if (buf->result < 0)
            return POLLERR;
        return POLLIN | POLLRDNORM;
    }
    return 0;
}

static struct page *siusb_vma_nopage(struct vm_area_struct *vma, unsigned long address, int unused)
{
    struct siusb *s = vma->vm_private_data;
    unsigned page_offset;   /* Faulting page number in this VMA */
    unsigned buf_nr;        /* Buffer number, the page belongs to */
    unsigned page_nr;       /* Page number in that buffer */
    void *pageptr;
    struct page *page;

    page_offset = vma->vm_pgoff + (address - vma->vm_start) / PAGE_SIZE;

    if (page_offset >= num_bufs * siusb_pages_per_buf)  /* Out of range */
        return NOPAGE_SIGBUS;

    buf_nr = page_offset / siusb_pages_per_buf;
    page_nr = page_offset % siusb_pages_per_buf;

    pageptr = (void *)s->bufs[buf_nr].pages[page_nr];
    page = virt_to_page(pageptr);

    get_page(page);
    return page;
}

static struct vm_operations_struct siusb_vm_ops = {
    .nopage =   siusb_vma_nopage,
};

static int siusb_mmap(struct file *file, struct vm_area_struct *vma)
{
    struct siusb *s = file->private_data;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    /* Don't do anything here, nopage() will fill the holes.
     */
    vma->vm_flags |= VM_RESERVED;
    vma->vm_ops = &siusb_vm_ops;
    vma->vm_private_data = s;

out:
    up(&s->sem);
    return result;
}

static int siusb_do_ioctl(struct inode *inode, struct file *file, unsigned cmd, void *arg)
{
    struct siusb *s = file->private_data;
    int result;

    result = down_interruptible(&s->sem);
    if (result < 0)
        return result;

    /* Verify that the camera wasn't unplugged.
     */
    if (!s->present) {
        result = -ENODEV;
        goto out;
    }

    switch (cmd) {
        case VIDIOCGCAP:
        {
            struct video_capability *cap = arg;

            strcpy(cap->name, s->version);
            cap->type = VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE | VID_TYPE_MONOCHROME;
            cap->channels = 1;
            cap->audios = 0;
            cap->maxwidth = SIUSB_MAX_WIDTH;
            cap->maxheight = SIUSB_MAX_HEIGHT;
            cap->minwidth = SIUSB_MIN_WIDTH;
            cap->minheight = SIUSB_MIN_HEIGHT;
            break;
        }

        case VIDIOCGCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0) {
                result = -EINVAL;
                break;
            }
            strcpy(chan->name, "Camera");
            chan->tuners = 0;
            chan->flags = 0;
            chan->type = VIDEO_TYPE_CAMERA;
            chan->norm = VIDEO_MODE_AUTO;
            break;
        }

        case VIDIOCSCHAN:
        {
            struct video_channel *chan = arg;

            if (chan->channel != 0 || chan->norm != VIDEO_MODE_AUTO)
                result = -EINVAL;
            break;
        }

        case VIDIOCGPICT:
        {
            struct video_picture *pict = arg;

            pict->brightness = s->exposure << 4;
            pict->hue = 32768;
            pict->colour = 32768;
            pict->contrast = 32768;
            pict->whiteness = 32768;
            pict->depth = 8;
            pict->palette = VIDEO_PALETTE_GREY;
            break;
        }

        case VIDIOCSPICT:
        {
            struct video_picture *pict = arg;

            if (pict->depth != 8 || pict->palette != VIDEO_PALETTE_GREY)
                result = -EINVAL;

            siusb_flush_captq(s);
            result = siusb_set_exposure(s, pict->brightness >> 4);
            break;
        }

        case VIDIOCGWIN:
        {
            struct video_window *win = arg;

            win->x = 0;
            win->y = 0;
            win->width = s->w;
            win->height = s->h;
            win->chromakey = 0;
            win->flags = 0;
            win->clips = NULL;
            win->clipcount = 0;
            break;
        }

        case VIDIOCSWIN:
        {
            struct video_window *win = arg;

            if (win->clips != NULL || win->clipcount != 0) {
                result = -EINVAL;
                break;
            }

            if (s->w != win->width || s->h != win->height)
                result = siusb_set_size(s, (SIUSB_MAX_WIDTH - win->width) / 2,
                                           (SIUSB_MAX_HEIGHT - win->height) / 2,
                                           win->width, win->height);
            break;
        }

        case VIDIOCGCAPTURE:
        {
            struct video_capture *capt = arg;

            capt->x = s->x;
            capt->y = s->y;
            capt->width = s->w;
            capt->height = s->h;
            capt->decimation = 0;
            capt->flags = 0;
            break;
        }

        case VIDIOCSCAPTURE:
        {
            struct video_capture *capt = arg;

            if (capt->decimation != 0 || capt->flags != 0) {
                result = -EINVAL;
                break;
            }

            if (s->x != capt->x || s->y != capt->y || s->w != capt->width || s->h != capt->height)
                result = siusb_set_size(s, capt->x, capt->y, capt->width, capt->height);
            break;
        }

        case VIDIOCGMBUF:
        {
            struct video_mbuf *mbuf = arg;
            unsigned i;

            mbuf->size = num_bufs * SIUSB_MAX_SIZE;
            mbuf->frames = num_bufs;
            for (i = 0; i < num_bufs; i++)
                mbuf->offsets[i] = i * SIUSB_MAX_SIZE;
            break;
        }

        case VIDIOCMCAPTURE:
        {
            struct video_mmap *mmap = arg;

            if (mmap->frame >= num_bufs) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad frame=%u\n", mmap->frame);
                result = -EINVAL;
                break;
            }

            if (mmap->format != VIDEO_PALETTE_GREY) {
                TRACE1("siusb: VIDIOCMCAPTURE: bad format=%u\n", mmap->format);
                result = -EINVAL;
                break;
            }

            /* TODO: we could test for busy buffer before resizing (as it was to be).
             */

            if (s->w != mmap->width || s->h != mmap->height) {
                result = siusb_set_size(s, (SIUSB_MAX_WIDTH - mmap->width) / 2,
                                           (SIUSB_MAX_HEIGHT - mmap->height) / 2,
                                           mmap->width, mmap->height);
                if (result < 0)
                    break;
            }

            result = siusb_capture_request(s, &s->bufs[mmap->frame]);
            break;
        }

        case VIDIOCSYNC:
        {
            unsigned *frame = arg;

            if (*frame >= num_bufs) {
                result = -EINVAL;
                break;
            }

            result = siusb_readout_wait(s, &s->bufs[*frame]);
            if (result < 0)
                break;

            /* TODO: what should return VIDIOCSYNC if there is frame capture error?
             *
             * We return -EINVAL on bad arg
             *           -EINTR if signaled (probably)
             * These two mean the frame is still NOT finished and the user should reinvoke
             * VIDIOCSYNC.
             *
             * The other case is the frame is finished, but with an error -- we return
             * the error code!
             *
             * So on error return how does the user find out if the VIDIOCSYNC call must
             * be reinvoked?
             */
            result = s->bufs[*frame].result;
            break;
        }

        case VIDIOCGUNIT:
        {
            struct video_unit *unit = arg;

            unit->video = s->vdev->minor;
            unit->vbi = VIDEO_NO_UNIT;
            unit->radio = VIDEO_NO_UNIT;
            unit->audio = VIDEO_NO_UNIT;
            unit->teletext = VIDEO_NO_UNIT;
            break;
        }

        /* No overlay.
         */
        case VIDIOCCAPTURE:
        case VIDIOCGFBUF:
        case VIDIOCSFBUF:
        case VIDIOCKEY:

        /* No tuner.
         */
        case VIDIOCGTUNER:
        case VIDIOCSTUNER:
        case VIDIOCGFREQ:
        case VIDIOCSFREQ:

        /* No audio.
         */
        case VIDIOCGAUDIO:
        case VIDIOCSAUDIO:
            result = -EINVAL;
            break;

        case SIUSB_IOC_GET_CLOCK:
        {
            unsigned *clk = arg;

            result = siusb_get_clock(s, clk);
            break;
        }

        case SIUSB_IOC_SET_CLOCK:
        {
            unsigned *clk = arg;

            siusb_flush_captq(s);
            result = siusb_set_clock(s, *clk);
            break;
        }

        case SIUSB_IOC_GET_SHUTTER_MODE:
        {
            unsigned *mode = arg;

            *mode = s->shutter_mode;
            break;
        }

        case SIUSB_IOC_SET_SHUTTER_MODE:
        {
            unsigned *mode = arg;

            siusb_flush_captq(s);
            result = siusb_set_shutter_mode(s, *mode);
            break;
        }

        case SIUSB_IOC_GET_EXPOSURE:
        {
            unsigned *exp = arg;

            *exp = s->exposure;
            break;
        }

        case SIUSB_IOC_SET_EXPOSURE:
        {
            unsigned *exp = arg;

            siusb_flush_captq(s);
            result = siusb_set_exposure(s, *exp);
            break;
        }

        case SIUSB_IOC_GET_GAIN:
        {
            unsigned *gain = arg;

            *gain = s->gain;
            break;
        }

        case SIUSB_IOC_SET_GAIN:
        {
            unsigned *gain = arg;

            siusb_flush_captq(s);
            result = siusb_set_gain(s, *gain);
            break;
        }

        case SIUSB_IOC_GET_BLACK_LEVEL:
        {
            unsigned *val = arg;

            *val = s->black_level;
            break;
        }

        case SIUSB_IOC_SET_BLACK_LEVEL:
        {
            unsigned *val = arg;

            siusb_flush_captq(s);
            result = siusb_set_black_level(s, *val);
            break;
        }

        case SIUSB_IOC_GET_BLACK_BALANCE:
        {
            unsigned *val = arg;

            *val = s->black_balance;
            break;
        }

        case SIUSB_IOC_SET_BLACK_BALANCE:
        {
            unsigned *val = arg;

            siusb_flush_captq(s);
            result = siusb_set_black_balance(s, *val);
            break;
        }

        case SIUSB_IOC_GET_TRIG_MODE:
        {
            unsigned *mode = arg;

            *mode = s->trig_mode;
            break;
        }

        case SIUSB_IOC_SET_TRIG_MODE:
        {
            unsigned *mode = arg;

            siusb_flush_captq(s);
            result = siusb_set_trig_mode(s, *mode);
            break;
        }

        default:
            result = -ENOIOCTLCMD;
            break;
    }

out:
    up(&s->sem);
    return result;
}

static int siusb_ioctl(struct inode *inode, struct file *file, unsigned cmd, unsigned long arg)
{
    return video_usercopy(inode, file, cmd, arg, siusb_do_ioctl);
}

static struct file_operations siusb_fops = {
    .owner =    THIS_MODULE,
    .open =     siusb_open,
    .release =  siusb_release,
    .read =     siusb_read,
    .poll =     siusb_poll,
    .mmap =     siusb_mmap,
    .ioctl =    siusb_ioctl,
    .llseek =   no_llseek,
};

static struct video_device siusb_vdev_template = {
    .owner =    THIS_MODULE,
    .type =     VID_TYPE_CAPTURE,
    .hardware = VID_HARDWARE_SI,
    .fops =     &siusb_fops,
    .release =  video_device_release,
};

static int siusb_v4l_register(struct siusb *s)
{
    int result;

    s->vdev = video_device_alloc();
    if (!s->vdev) {
        printk(KERN_ERR "siusb: video_device_alloc(): FAILED\n");
        return -ENOMEM;
    }

    *s->vdev = siusb_vdev_template;
#ifndef LINUX_24
    s->vdev->dev = &s->udev->dev;
#endif
    strcpy(s->vdev->name, s->version);
    video_set_drvdata(s->vdev, s);

    result = video_register_device(s->vdev, VFL_TYPE_GRABBER, -1);
    if (result < 0) {
        printk(KERN_ERR "siusb: video_register_device()=%d: FAILED\n", result);
        video_device_release(s->vdev);
        return result;
    }

    return 0;
}

static void siusb_v4l_unregister(struct siusb *s)
{
    video_unregister_device(s->vdev);
}


/*
 * Initialization
 */

static int siusb_reset_cam(struct siusb *s)
{
    int result;

    result = siusb_set_clock(s, SIUSB_CLK_20MHZ);           /* Clock */
    if (result < 0)
        return result;
    result = siusb_set_shutter_mode(s, SIUSB_SH_TRIG_TB3);  /* Shutter mode */
    if (result < 0)
        return result;
    result = siusb_set_exposure(s, 0x800);                  /* Exposure time */
    if (result < 0)
        return result;
    result = siusb_set_gain(s, SIUSB_GAIN_1_88);            /* Gain */
    if (result < 0)
        return result;
    result = siusb_set_black_level(s, 85);                  /* Black level */
    if (result < 0)
        return result;
    result = siusb_set_black_balance(s, 85);                /* Black balance */
    if (result < 0)
        return result;
    result = siusb_set_origin(s, 0, 0);                     /* Frame origin */
    if (result < 0)
        return result;
    result = siusb_set_width(s, SIUSB_MAX_WIDTH);           /* Frame width */
    if (result < 0)
        return result;
    result = siusb_set_height(s, SIUSB_MAX_HEIGHT);         /* Frame height */
    if (result < 0)
        return result;
    result = siusb_set_depth(s, SIUSB_DEPTH_8BIT);          /* Bit depth */
    if (result < 0)
        return result;
    result = siusb_set_trig_mode(s, SIUSB_TRIG_MODE_INT);   /* Trigger mode */
    if (result < 0)
        return result;
    return 0;
}

#ifdef LINUX_24
static void *siusb_probe(struct usb_device *udev, unsigned ifnum, const struct usb_device_id *id)
#else
static int siusb_probe(struct usb_interface *ifc, const struct usb_device_id *id)
#endif
{
    struct siusb *s;
#ifdef LINUX_24
    struct usb_interface *ifc;
#endif
    int result;
    struct usb_endpoint_descriptor *ep_desc;

    /* Allocate memory for our device state.
     */
    s = kmalloc(sizeof(*s), GFP_KERNEL);
    if (!s) {
        printk(KERN_ERR "siusb: %s: kmalloc() failed\n", __func__);
        result = -ENOMEM;
        goto err;
    }

    /* And initialize it.
     */
    memset(s, 0, sizeof(*s));
#ifdef LINUX_24
    s->udev = udev;
    ifc = usb_ifnum_to_if(udev, ifnum);
#else
    s->udev = interface_to_usbdev(ifc);
#endif
    s->present = 1;         /* Allow file operations */
    init_MUTEX(&s->sem);

    /* Do some sanity checks.
     */
    if (s->udev->descriptor.bNumConfigurations != 1) {
        printk(KERN_ERR "siusb: bNumConfigurations != 1\n");
        result = -ENODEV;
        goto err;
    }
#ifdef LINUX_24
    ep_desc = &ifc->altsetting[0].endpoint[0];
#else
    ep_desc = &ifc->altsetting[0].endpoint[0].desc;
#endif
    if (!(ep_desc->bEndpointAddress & USB_DIR_IN)
     || !((ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)) {
        printk(KERN_ERR "siusb: endpoint not bulk IN\n");
        result = -ENODEV;
        goto err;
    }
    if (ep_desc->wMaxPacketSize != SIUSB_PACKET_SIZE) {
        printk(KERN_ERR "siusb: wMaxPacketSize != " __stringify(SIUSB_PACKET_SIZE) "\n");
        result = -ENODEV;
        goto err;
    }

    s->bulk_ep_addr = ep_desc->bEndpointAddress;
    TRACE2("siusb: bulk_ep_addr=0x%02x\n", s->bulk_ep_addr);

    /* Reset camera.
     */
    result = siusb_reset_cam(s);
    if (result < 0)
        goto err;

    /* Read camera firmware version.
     */
    result = siusb_read_version(s, s->version);
    if (result < 0)
        goto err;

    /* Init capture queue.
     */
    INIT_LIST_HEAD(&s->captq);
    spin_lock_init(&s->captq_lock);

    result = siusb_v4l_register(s);
    if (result < 0)
        goto err;

    printk(KERN_INFO "siusb: %s attached\n", s->version);
#ifdef LINUX_24
    return s;
#else
    usb_set_intfdata(ifc, s);
    return 0;
#endif

err:
    kfree(s);
#ifdef LINUX_24
    return NULL;
#else
    return result;
#endif
}

#ifdef LINUX_24
static void siusb_disconnect(struct usb_device *udev, void *context)
#else
static void siusb_disconnect(struct usb_interface *ifc)
#endif
{
    struct siusb *s;

    down(&disconnect_sem);

#ifdef LINUX_24
    s = context;
#else
    s = usb_get_intfdata(ifc);
    usb_set_intfdata(ifc, NULL);
#endif

    down(&s->sem);
    s->present = 0;             /* Prevent file operations */
    siusb_v4l_unregister(s);
    up(&s->sem);

    /* If the device is opened, siusb_release() will clean this up.
     */
    if (s->open_count <= 0)
        kfree(s);

    up(&disconnect_sem);
    printk(KERN_INFO "siusb: disconnected\n");
}

static struct usb_device_id siusb_id_table[] = {
    {USB_DEVICE(0x0547, 0x1002)},
    { } /* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, siusb_id_table);

static struct usb_driver siusb_driver = {
    .owner =        THIS_MODULE,
    .name =         "siusb",
    .probe =        siusb_probe,
    .disconnect =   siusb_disconnect,
    .id_table =     siusb_id_table,
};

static int __init siusb_init(void)
{
    int result;

    BUG_ON(SIUSB_MAX_SIZE % PAGE_SIZE != 0);    /* We rely on this */
    siusb_pages_per_buf = SIUSB_MAX_SIZE / PAGE_SIZE;
    siusb_read_buf_nr = num_bufs - 1;

    result = usb_register(&siusb_driver);
    if (result < 0)
        return result;

    printk(KERN_INFO DRIVER_DESC " " DRIVER_VERSION "\n");
    return 0;
}

static void __exit siusb_exit(void)
{
    usb_deregister(&siusb_driver);
}

module_init(siusb_init);
module_exit(siusb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Proprietary");
